@@ -8,7 +8,7 @@
  work, as well as functions which operate on them.
 """
 from types import FunctionType
-from typing import Any, Callable, Dict, Generic, Tuple, TypeVar
+from typing import Any, Dict, Generic, TypeVar
 
 DataClass = Any  # type hint for variables that should be data class instances
 T = TypeVar('T')
@@ -65,7 +65,7 @@ class DataClassMeta(type):
                 del dict_[d]
             del dict_['__slots__']
         if options['init']:
-            dict_.setdefault('__init__', _generate_init(all_annotations, all_defaults, options['kwargs']))
+            dict_.setdefault('__new__', _generate_new(all_annotations, all_defaults, options['kwargs']))
         if options['repr']:
             dict_.setdefault('__repr__', _generated_repr)
         if options['eq']:
@@ -78,76 +78,36 @@ class DataClassMeta(type):
         return type.__new__(mcs, name, bases, dict_)
 
 
-def is_dataclass(obj: Any) -> bool:
-    """Return True if the given object is a data class as implemented in this package, otherwise False."""
-    return getattr(obj, '__metaclass__', None) is DataClassMeta
+def _generate_new(class_annotations: Dict[str, Any], defaults: Dict[str, Any], gen_kwargs: bool) -> FunctionType:
+    """Generate and return a __new__ method for a data class which has as parameters all fields of the data class.
+    When the data class is initialised, arguments to this function are applied to the fields of the new instance. Using
+    __new__ frees up __init__, allowing it to be defined by the user to perform additional, custom initialisation."""
+    user_init = '__init__' in defaults
 
-
-def is_dataclass_instance(obj: Any) -> bool:
-    """Return True if the given object is an instance of a data class, otherwise False."""
-    return is_dataclass(obj) and type(obj) is not DataClassMeta
-
-
-def fields(dataclass: DataClass, internals=False) -> Dict[str, Any]:
-    """Return a dict of `dataclass`'s fields and their values. `internals` selects whether to include internal fields.
-    A field is defined as a class-level variable with a type annotation."""
-    assert is_dataclass_instance(dataclass)
-    return {f: getattr(dataclass, f) for f in _filter_annotations(dataclass.__annotations__, internals)}
-
-
-def as_dict(dataclass: DataClass, dict_factory=dict) -> Dict[str, Any]:
-    """Recursively create a dict of a dataclass instance's fields and their values.
-    This function is recursively called on data classes, named tuples and iterables."""
-    assert is_dataclass_instance(dataclass)
-    return _recurse_structure(dataclass, dict_factory)
-
-
-def as_tuple(dataclass: DataClass) -> Tuple:
-    """Recursively create a tuple of the values of a dataclass instance's fields, in definition order.
-    This function is recursively called on data classes, named tuples and iterables."""
-    assert is_dataclass_instance(dataclass)
-    return _recurse_structure(dataclass, lambda k_v: tuple(v for k, v in k_v))
-
-
-def _generate_init(class_annotations: Dict[str, Any], defaults: Dict[str, Any], gen_kwargs: bool) -> FunctionType:
-    """Generate and return an __init__ method for a data class."""
-    arguments = [str(a) for a in class_annotations if a not in defaults]
+    arguments = [a for a in class_annotations if a not in defaults]
     default_arguments = [f'{a}={a}' for a in class_annotations if a in defaults]
-    kwargs = ['**kwargs' if gen_kwargs else '']
-    assignments = [f'object.__setattr__(self, {k!r}, {k}.copy() if hasattr({k}, "copy") else {k})' for k
-                   in class_annotations]
+    args = ['*args'] if user_init else []
+    kwargs = ['**kwargs'] if gen_kwargs or user_init else []
 
-    signature = f'lambda self, {", ".join(arguments + default_arguments + kwargs)}: '
-    body = ' or '.join(assignments) or 'None'
-    function = eval(signature + body, {}, defaults)  # an empty globals ensures defaults is not polluted with builtins
+    parameters = ', '.join(arguments + default_arguments + args + kwargs)
+
+    assignments = [f'object.__setattr__(self, {k!r}, {k}.copy() if hasattr({k}, "copy") else {k})'
+                   for k in class_annotations]
+
+    signature = f'def __new__(cls, {parameters}):'
+    statements = [f'self = object.__new__(cls)', *assignments, f'self.__init__({parameters})', 'return self']
+
+    exec('\n\t'.join([signature, *statements]), {}, defaults)
+    function = defaults.pop('__new__')
     function.__annotations__ = class_annotations
     return function
 
 
-# method overrides which are constant for all data classes
+# method implementations which are constant for all data classes
+from .functions import fields, as_tuple
+
 _generated_eq = lambda self, other: as_tuple(self) == as_tuple(other)
 _generated_iter = lambda self: iter(as_tuple(self))
 _generated_repr = lambda self: f'{self.__class__.__name__}(' \
         f'{", ".join(f"{f}={v!r}" for f, v in fields(self, not self.__dataclass__["hide_internals"]).items())})'
 _generated_attr = lambda *args: exec('raise AttributeError("Frozen class")')
-
-
-def _filter_annotations(annotations: Dict[str, Any], internals: bool) -> Dict[str, Any]:
-    """Filter an annotations dict for to remove or keep internal fields."""
-    return annotations if internals else {f: a for f, a in annotations.items()
-                                          if not f.startswith('_') and not Internal.is_internal(a)}
-
-
-def _recurse_structure(var: Any, iter_proc: Callable) -> Any:
-    """Recursively convert an arbitrarily nested structure beginning at `var`, copying and processing any iterables
-    encountered with `iter_proc`."""
-    if is_dataclass(var):
-        var = fields(var, internals=True)
-    if hasattr(var, '_asdict'):  # handle named tuples
-        # noinspection PyCallingNonCallable, PyProtectedMember
-        var = var._asdict()
-    if isinstance(var, dict):
-        return iter_proc((_recurse_structure(k, iter_proc), _recurse_structure(v, iter_proc)) for k, v in var.items())
-    if isinstance(var, (list, tuple)):
-        return type(var)(_recurse_structure(e, iter_proc) for e in var)
-    return var
